.386p

;-----------------------------------------;
;    This asm module allows programmer    ;
; to use all computer memory (up to 4GB). ;
;                                         ;
; Works in pure DOS.   Needs Real Mode of ;
; processor.   Don't try to execute under ;
; Windows. You'll get nothing. Better     ;
; load previous DOS version or boot using ;
; DOS system diskette.                    ;
;                                         ;
;    Program doesn't work under memory    ;
;   managers such as Himem.sys, Emm.exe,  ;
;             Qemm.exe etc.               ;
;                                         ;
; Written by Alexei. A. Frounze  (c) 1999 ;
; E-mail:     alexfru@chat.ru             ;
; Homepage:   http://www.chat.ru/~alexfru ;
;-----------------------------------------;

Public  Enable_A20, Disable_A20
Public  Check_Safety, XMS_Avail, PMode
Public  MemRB, MemRW, MemRL, MemWB, MemWW, MemWL

Code    Segment Para Public Use16
        Assume  CS:Code

;--------------------------- A20 line switching ----------------------------

Enable_A20 Proc Pascal Near
        Mov     AL, 0D1h
        Out     64h, AL
        Mov     AL, 0DFh
        Out     60h, AL
        Ret
Enable_A20 EndP

Disable_A20 Proc Pascal Near
        Mov     AL, 0D1h
        Out     64h, AL
        Mov     AL, 0DDh
        Out     60h, AL
        Ret
Disable_A20 EndP

;----------- Checking processor mode and looking for Himem.sys -------------

Check_Safety Proc Pascal Near
        SMSW    CS:XXX
        Mov     AX, CS:XXX
        And     AX, 1
        JNZ     @@CSQ           ; 1 - Processor is in Protected Mode
        Mov     AX, 4300h
        Int     2Fh
        Shr     AX, 6
        And     AX, 2
        JNZ     @@CSQ           ; 2 - Himem.sys is installed
                                ; 0 - OK
@@CSQ:
        Ret
XXX     DW      ?
Check_Safety EndP

;------------------------- Getting XMS ammount -----------------------------

Read_CMOS Proc Pascal Near
        Out     70h, AL   ; CMOS address should be stored in the AL register
        Jmp     @@RCW     ; Little delay
@@RCW:
        In      AL, 71h   ; AL = value
        Ret
Read_CMOS EndP

XMS_Avail Proc Pascal Near
        Mov     AL, 31h   ; 31h = hi byte address
        Call    Read_CMOS
        Mov     AH, AL
        Mov     AL, 30h   ; 30h = low byte address
        Call    Read_CMOS
        Ret               ; AX = XMS installed above 1MB (in Kilobytes)
XMS_Avail EndP

;-------- Protected mode intialization and creation of 4GB segment ---------

PMode   Proc Pascal Near
        Xor     EAX, EAX
        Mov     AX, CS
        Shl     EAX, 4
        LEA     EDX, GDT
        Add     EAX, EDX
        Mov     DWord Ptr CS:GDTR+2, EAX; Fill in GDTR with physical address
                                        ; of Global Descriptor Table
        Push    DS
        CLI
        LGDT    FWord Ptr CS:GDTR       ; Load Global Descriptor Table
        Mov     EAX, 1
        Mov     CR0, EAX                ; Set Protected Mode
        Mov     AX, 8
        Mov     DS, AX
        Mov     ES, AX
        Mov     FS, AX
        Mov     GS, AX                  ; All segment registers are loaded
                                        ; with 4GB segment's selector
        Xor     EAX, EAX
        Mov     CR0, EAX                ; Set Real Mode
        STI
        Pop     DS

;----------------------------------------------------------;
; Now you can access all the memory by putting zero to a   ;
; segment register and 32bit physical address to an index  ;
; register.                                                ;
;                     -= Simple example =-                 ;
; Xor AX, AX                                               ;
; Mov ES, AX                  ; zero the ES register       ;
; Mov EDI, 1024*1024+65536    ; 1st byte we can use freely ;
; Mov BL, Byte Ptr ES:[EDI]   ; read byte                  ;
; Mov DWord Ptr ES:[EDI], EDX ; write dword                ;
;----------------------------------------------------------;

        Ret

GDT     DQ      0, 8F92000000FFFFh      ; "0" & "4GB" Descriptors
GDTR    DW      16, 0, 0
PMode   EndP

;------------------------ Memory reading functions -------------------------

MemRB   Proc Pascal Near                ;Offs : DWord
        Push    BP
        Mov     BP, SP
        Push    DS
        Xor     AX, AX
        Mov     DS, AX
        Mov     EBX, DWord Ptr [BP+4]
        Mov     AL, [EBX]
        Pop     DS
        Pop     BP
        Ret     4
MemRB   EndP

MemRW   Proc Pascal Near                ;Offs : DWord
        Push    BP
        Mov     BP, SP
        Push    DS
        Xor     AX, AX
        Mov     DS, AX
        Mov     EBX, DWord Ptr [BP+4]
        Mov     AX, [EBX]
        Pop     DS
        Pop     BP
        Ret     4
MemRW   EndP

MemRL   Proc Pascal Near                ;Offs : DWord
        Push    BP
        Mov     BP, SP
        Push    DS
        Xor     AX, AX
        Mov     DS, AX
        Mov     EBX, DWord Ptr [BP+4]
        Mov     EAX, [EBX]
        Shld    EDX, EAX, 16
        Pop     DS
        Pop     BP
        Ret     4
MemRL   EndP

;------------------------ Memory writing functions -------------------------

MemWB   Proc Pascal Near                ;Offs : DWord, Value : Byte
        Push    BP
        Mov     BP, SP
        Push    DS
        Xor     AX, AX
        Mov     DS, AX
        Mov     EBX, DWord Ptr [BP+6]
        Mov     AL, Byte Ptr [BP+4]
        Mov     [EBX], AL
        Pop     DS
        Pop     BP
        Ret     6
MemWB   EndP

MemWW   Proc Pascal Near                ;Offs : DWord, Value : Word
        Push    BP
        Mov     BP, SP
        Push    DS
        Xor     AX, AX
        Mov     DS, AX
        Mov     EBX, DWord Ptr [BP+6]
        Mov     AX, Word Ptr [BP+4]
        Mov     [EBX], AX
        Pop     DS
        Pop     BP
        Ret     6
MemWW   EndP

MemWL   Proc Pascal Near                ;Offs : DWord, Value : DWord
        Push    BP
        Mov     BP, SP
        Push    DS
        Xor     AX, AX
        Mov     DS, AX
        Mov     EBX, DWord Ptr [BP+8]
        Mov     EAX, DWord Ptr [BP+4]
        Mov     [EBX], EAX
        Pop     DS
        Pop     BP
        Ret     8
MemWL   EndP

Code    EndS

End
